bits	16
org		0x0000

jmp		word start

[map	symbols debug/boot.map]

; Includes
%include "boot/bpb.s"

start:
	; Set registers
	mov		ax, 0x07C0
	mov		ds, ax
	mov		es, ax
	mov		fs, ax
	mov		gs, ax
	
	mov		si, MSG_LOADING
	call	print_msg
	
	; The purpose of this bootloader is to load the second stage bootloader
	; :: KRNLDR.BIN (krnldr.bin) which will, in turn, boot into our kernel
	; The purpose of KRNLDR.BIN is to setup the A20 gate (access 4GB of memory),
	; setup the initial GDT (Global Descriptor Table), and setup the Multiboot Information
	; including the Memory Map and the various flags / drive number for our kernel
	; to use when setting up it's PMM (Physical Memory Manager) and VMM (Virtual Memory Manager)
	
	; To load KRNLDR.BIN, first we need to make sure it exists. The file will be store in
	; the Root Directory of the bootloader
load_root_dir:
	mov		ax, 32
	mul		word [RootDirEntries]
	div		word [BytesPerSector]
	xchg	ax, cx
	
	movzx	ax, byte [NumberOfFATs]
	mul		word [SectorsPerFAT]
	add		ax, word [ReservedSectors]
	
	mov		word [DataSector], ax
	add		word [DataSector], cx
	
	mov		bx, 0x0200
	call	read_sectors
	
; Now we have the root directory stored at 0x7E00
; We want to now search for the file in the root directory
find_kernel_loader:
	mov		cx, word [RootDirEntries]
	mov		di, 0x0200
	
	.loop:
		push	cx
		mov		cx, 11				; Image size is 11 bytes
		mov		si, KRNLDR_NAME		; DS:SI = Name to find, ES:DI = The actual location
		
		push	di
		rep		cmpsb				; Check whether or not this is the file name
		pop		di
		
		je		load_fat			; If we found the file, load the FAT (which contains cluster info)

		pop		cx
		add		di, 32				; Go to next root directory entry
		loop	.loop				; Search again
		
		jmp		failure				; If we never found the file, bail out
		
load_fat:
	mov		dx, [di + 0x001A]		; Index into the FAT and get the starting cluster
	mov		word [Cluster], dx

	; Get the size of the FAT
	xor		cx, cx
	movzx	ax, byte [NumberOfFATs]
	mul		word [SectorsPerFAT]
	mov		cx, ax
	
	; Get the start of the FAT
	mov		ax, word [ReservedSectors]
	
	mov		bx, 0x0200
	call	read_sectors	
	
	mov		ax, 0x0050
	mov		es, ax
	mov		bx, 0x0000
	push	bx

; Now that we have the FAT loaded at 0x07C0:0x0200, we can search through the FAT 
; for the clusters that make up KRNLDR.BIN
load_image:
	mov		ax, word [Cluster]	
	pop		bx
	
	call	chs_cluster_to_lba				; AX = Start of cluster

	movzx	cx, byte [SectorsPerCluster]	; CX = Size of cluster (1 sector)
	call	read_sectors	
	push	bx
	
	; Get the next cluster
	mov		ax, word [Cluster]
	mov		cx, ax
	mov		dx, ax
	shr		dx, 0x0001
	add		cx, dx
	mov		bx, 0x0200
	add		bx, cx
	mov		dx, word [bx]
	
	test	ax, 1
	jnz		.odd_cluster
	
	.even_cluster:
		and		dx, 0000111111111111b
		jmp		.done
		
	.odd_cluster:
		shr		dx, 4
		
	.done:
		mov		word [Cluster], dx
		cmp		dx, 0x0FF0
		jb		load_image
	
done:
	mov		si, MSG_BOOTING
	call	print_msg
	
	movzx	dx, byte [DriveNumber]
	
	jmp		0x0050:0x0000
	
	cli
	hlt
	
failure:
	mov		si, MSG_FAILURE
	call	print_msg

	cli
	hlt

; ------------	
; read_sectors
; Read a number of sectors to the disk
; In:
;  AX => The LBA sector to read
;  CX => The number of sectors to read
;  ES:BX => The location to read the sectors that we want to
; ------------
read_sectors:
	.start:
		mov		di, 5		; We will try 5 times to read the sectors
		
		push	ax
		push	bx
		push	cx
		
		call	lba_to_chs	; Convert the LBA address to CHS address info
		
	.loop:	
		mov		ah, 0x02	; BIOS Function INT 0x13, AH 0x2
		mov		al, 0x01	; We will read 1 sector at a time
		mov		ch, byte [AbsoluteTrack]
		mov		cl, byte [AbsoluteSector]
		mov		dh, byte [AbsoluteHead]
		mov		dl, byte [DriveNumber]
		int		0x13
		jnc		.success
		
		xor		ax, ax
		int		0x13
		
		pop		cx
		pop		bx
		pop		ax
		dec		di
		
		jnz		.loop
		
		xor		ax, ax
		int		0x16
		int		0x19
		
	.success:
		mov		ah, 0x0e
		mov		al, '*'
		int		0x10
	
		pop		cx
		pop		bx
		pop		ax
		
		add		bx, word [BytesPerSector]
		inc		ax
		loop	.start
		
		ret
		

; ---------- 
; lba_to_chs
; Extract CHS information from LBA Sector Address
; In:
;  AX => The LBA Sector Address
; Out:
;  AbsoluteSector => The sector
;  AbsoluteHead => The head
;  AbsoluteTrack => The track
; ----------
lba_to_chs:
	xor		dx, dx
	div		word [SectorsPerTrack]
	inc		dl
	mov		byte [AbsoluteSector], dl
	
	xor		dx, dx
	div		word [NumberOfHeads]
	mov		byte [AbsoluteHead], dl
	mov		byte [AbsoluteTrack], al	
	ret
	

; chs_cluster_to_lba
; Converts a CHS Cluster Address to LBA address
; In:
;  AX => CHS Cluster Address
; Out:
;  AX => LBA Sector Address	
chs_cluster_to_lba:
	sub		ax, 2
	mul		byte [SectorsPerCluster]
	
	add		ax, word [DataSector]
	ret
	
; ---------
; print_msg
; Prints a message to the screen
; In:
;  ES:DI => The message to print
; ---------
print_msg:
	lodsb
	cmp		al, 0x00
	je		.done
	mov		ah, 0x0e
	int		0x10
	jmp		print_msg
	
	.done:
		ret
	
; Strings
MSG_LOADING		db "Starting DarkBoot...",10,13,0
MSG_FAILURE		db 10,13,"[DARKBOOT]: Failed to find KRNLDR.BIN",10,13,0
MSG_BOOTING		db 10,13,'[DARKBOOT]: Booting into KRNLDR.BIN',10,13,0

KRNLDR_NAME		db "KRNLDR  BIN"

; Variables
AbsoluteSector	db 0x00
AbsoluteHead	db 0x00
AbsoluteTrack	db 0x00

DataSector		dw 0x0000
Cluster			dw 0x0000	

times 510-($-$$) db 0
dw 0xAA55