%ifndef FAT12_S
%define FAT12_S

; BIOS Parameter Block Info (Defined in BPB.s)
%define BytesPerSector		0x7C00 + 11
%define SectorsPerCluster	0x7C00 + 13
%define ReservedSectors		0x7C00 + 14
%define NumberOfFATs		0x7C00 + 16
%define RootDirEntries		0x7C00 + 17
%define SectorsPerFAT		0x7C00 + 22
%define SectorsPerTrack		0x7C00 + 24
%define NumberOfHeads		0x7C00 + 26
%define DriveNumber			0x7C00 + 36

; Definitions
%define ROOT_SEG	0x02C0
%define ROOT_OFF	0x2C00
%define FAT_SEG		0x02E0
%define FAT_OFF		0x2E00

; Variables
DataSector		dw 0x0000
Cluster			dw 0x0000

AbsoluteSector	db 0x00
AbsoluteHead	db 0x00
AbsoluteTrack	db 0x00

; -------------
; load_root_dir
; Loads the root directory to 0x2C00
; -------------
load_root_dir:
	pusha
	push	es
	
	; Get the size of the root director
	xor		cx, cx
	mov		ax, 32
	mul		word [RootDirEntries]
	div		word [BytesPerSector]
	xchg	ax, cx
	
	; Get the start of the root directory
	movzx	ax, byte [NumberOfFATs]
	mul		word [SectorsPerFAT]
	add		ax, word [ReservedSectors]
	
	mov		word [DataSector], ax
	add		word [DataSector], cx
	
	mov		bx, ROOT_SEG
	mov		es, bx
	xor		bx, bx
	
	call	read_sectors
	
	pop		es
	popa
	ret

; --------
; load_fat
; Loads the FAT into memory at 0x2E00	
; --------
load_fat:
	pusha
	push	es
	
	; Get the size of the FATs
	xor		cx, cx
	movzx	ax, byte [NumberOfFATs]
	mul		word [SectorsPerFAT]
	mov		cx, ax
	
	; Get the start of the FAT
	mov		ax, word [ReservedSectors]
	
	mov		bx, FAT_SEG
	mov		es, bx
	xor		bx, bx
	
	call	read_sectors
	
	pop		es
	popa
	ret
	
; ---------
; find_file
; Finds a file in the disk
; In:
;  DS:SI => The file to find
; Out:
;  AX => -1 on error, file index number in root dir on success
; ---------
find_file:
	push	cx
	push	dx
	push	bx
	mov		bx, si
	
	; Brows the root directory for the binary image
	mov		cx, word [RootDirEntries]
	mov		di, ROOT_OFF
	cld
	
	.loop:
		push	cx
		mov		cx, 11
		mov		si, bx
		
		push	di
		rep		cmpsb
		pop		di
		
		je		.found
		
		pop		cx
		add		di, 32
		loop	.loop
		
	.not_found:
		pop		bx
		pop		dx
		pop		cx
		mov		ax, -1
		ret
		
	.found:
		pop		ax
		pop		bx
		pop		dx
		pop		cx
		ret
	
; ---------
; load_file
; Loads a file from the disk and stores it in memory
; In:
;  DS:SI => The file to load
;  BX:BP => The location to load the file to
; Out:
;  AX => -1 on error, 0 on success
;  CX => Number of sectors loaded
; ---------
load_file:
	xor		ecx, ecx
	push	ecx
	
	.find:
		push	bx
		push	bp
		
		call	find_file
		
		cmp		ax, -1
		jne		.load_image_pre
		
		pop		bp
		pop		bx
		pop		ecx
		mov		ax, -1
		ret
		
	.load_image_pre:
		sub		edi, ROOT_OFF		; Get the index number from Root Dir
		sub		eax, ROOT_OFF		
		
		; Get the starting cluster
		push	word ROOT_SEG
		pop		es
		mov		dx, word [es:di + 0x001A]
		mov		word [Cluster], dx
		
		pop		bx
		pop		es
		push	bx
		push	es
		
		call	load_fat
		
	.load_image:
		mov		ax, word [Cluster]
		pop		es
		pop		bx
		
		call	chs_cluster_to_lba
		
		movzx	cx, byte [SectorsPerCluster]
		call	read_sectors
		
		pop		ecx
		inc		ecx
		push	ecx
		
		push	bx
		push	es
		
		mov		ax, FAT_SEG
		mov		es, ax
		xor		bx, bx
		
		; Compute the next cluster
		mov		ax, word [Cluster]
		mov		cx, ax
		mov		dx, ax
		shr		dx, 1
		add		cx, dx
		
		mov		bx, 0
		add		bx, cx
		mov		dx, word [es:bx]
		test	ax, 1
		jnz		.odd_cluster
		
		.even_cluster:
			and		dx, 0000111111111111b
			jmp		.done
			
		.odd_cluster:
			shr		dx, 4
			
		.done:
			mov		word [Cluster], dx
			cmp		dx, 0x0FF0
			jb		.load_image
			
done:
	pop		es
	pop		bx
	pop		ecx
	xor		ax, ax
	ret
	
; ------------
; read_sectors
; Reads sectors into memory
; In:
;  AX => The LBA Address to start reading from
;  CX => The number of sectors to read
;  ES:BX => The buffer to read sectors to
; ------------
read_sectors:
	.start:
		mov		di, 5		; 5 tries to read a sector
		
	.loop:
		push	ax
		push	bx
		push	cx
		
		call	lba_to_chs
		
		mov		ah, 0x02
		mov		al, 0x01
		mov		ch, byte [AbsoluteTrack]
		mov		cl, byte [AbsoluteSector]
		mov		dh, byte [AbsoluteHead]
		mov		dl, byte [DriveNumber]
		
		int		0x13
		jnc		.success
		
		; On error, reset the drive and try again until DI = 0
		xor		ax, ax
		int		0x13
		
		pop		cx
		pop		bx
		pop		ax
		dec		di
		
		jnz		.loop
		
		xor		ax, ax
		int		0x16
		int		0x19
		
	.success:
		pop		cx
		pop		bx
		pop		ax
		
		add		bx, word [BytesPerSector]
		inc		ax
		
		loop	.start
		ret
		
; ----------
; lba_to_chs
; Converts LBA Sector Address to CHS Info
; In:
;  AX => LBA Sector Address
; Out:
;  AbsoluteSector => Absolute Sector
;  AbsoluteHead => Absolute Head
;  AbsoluteTrack => Absolute Track
; ----------
lba_to_chs:
	xor		dx, dx
	div		word [SectorsPerTrack]
	inc		dl
	mov		byte [AbsoluteSector], dl
	
	xor		dx, dx
	div		word [NumberOfHeads]
	mov		byte [AbsoluteHead], dl
	mov		byte [AbsoluteTrack], al	
	ret
	

; chs_cluster_to_lba
; Converts a CHS Cluster Address to LBA address
; In:
;  AX => CHS Cluster Address
; Out:
;  AX => LBA Sector Address	
chs_cluster_to_lba:
	sub		ax, 2
	mul		byte [SectorsPerCluster]
	
	add		ax, word [DataSector]
	ret

%endif