%ifndef MEMORY_S
%define MEMORY_S

; A memory map entry
struc mem_map_entry
	.entry_base		resq 1
	.entry_length	resq 1
	.entry_type		resd 1
	.acpi_field		resd 1
endstruc

; get_memory_map
; Loads the memory map from the operating system
; In:
;  ES:DI => The destination buffer for the various entries
; Out:
;  BP => The entry count (trashes all registers except ESI)
get_memory_map:
	xor		ebx, ebx
	xor		bp, bp
	mov		edx, 0x534D4150			; 'SMAP' in hex
	mov		eax, 0xE820
	
	; Force a valid ACPI 3.X entry
	mov		[es:di + mem_map_entry.acpi_field], dword 1
	mov		ecx, 24
	int		0x15
	
	jc		short .failed
	mov		edx, 0x534D4150			; Some BIOSes apparently trash this register
	cmp		eax, edx				; If the fetch was successful, EAX = EDX = 'SMAP'
	jne		short .failed
	
	test 	ebx, ebx				; EBX = 0 implies that the list is only 1 entry long
	je		short .failed
	jmp		short .entry
	
	.loop:
		mov		eax, 0xE820
		mov		[es:di + mem_map_entry.acpi_field], dword 1
		mov		ecx, 24
		int		0x15
		jc		short .finished
		mov		edx, 0x534D4150
	
	.entry:
		jcxz	.skip_entry			; If CX = 0, skip this entry
		cmp		cl, 20				; If length = 20, ACPI 3.X is null (valid one is 24 bytes)
		jbe		short .not_ext		; If the length <= 20, it is not an extended entry
		
		test	byte [es:di + mem_map_entry.acpi_field], 1	; Otherwise, see if the ACPI 3.X field was set (if it was changed from 1)
		je		short .skip_entry
		
	.not_ext:
		mov		ecx, [es:di + mem_map_entry.entry_length]		; Get the lower dword of the length
		or		ecx, [es:di + mem_map_entry.entry_length + 4]	; OR it with the upper dword
		jz		.skip_entry
		
		inc		bp					; Otherwise, increment BP
		add		di, 24				; and add 24 to DI for next entry
		
	.skip_entry:
		test	ebx, ebx
		jne		short .loop			; If EBX resets to zero, we are done
		
	.finished:
		clc							; Clear the carry flag and return
		ret
		
	.failed:
		stc							; Set the carry flag and return
		ret

; ---------------
; get_memory_size
; Gets the amount of memory (in KB) below 1 MB
; Out:
;  AX => The size (in KB), -1 on error
; ---------------
get_memory_size:
	clc
	int		0x12
	
	jc		.error
	test	ax, ax			; Size is 0
	je		.error
	
	cmp		ah, 0x86		; Unsupported function
	je		.error
	cmp		ah, 0x80		; Invalid command
	je		.error
	
	ret
	
	.error:
		mov		ax, -1
		ret
		
; -------------------
; get_ext_memory_size
; Gets the amount of memory up to 15 MB or 64 MB depending on BIOS
; Out:
;  AX => KB of extended memory (-1 on error)
; -------------------
get_ext_memory_size:
	clc
	mov		ah, 0x88
	int		0x15
	je		.error
	
	test	ax, ax
	je		.error
	
	cmp		ah, 0x86
	je		.error
	cmp		ah, 0x80
	je		.error
	
	ret
	
	.error:
		mov		ax, -1
		ret

; ---------------------
; get_memory_size_16_mb
; Gets the amount of memory in KB above 1 MB
; Out:
;  AX => KB between 1 MB and 16 MB (-1 on error)
;  BX => Number of 64K blocks above 16 MB (0 on error)
; ---------------------
get_memory_size_16_mb:
	clc
	
	push	ecx
	push	edx
	
	xor		ecx, ecx
	xor		edx, edx
	
	mov		ax, 0xE801
	int		0x15
	jc		.error
	
	cmp		ah, 0x86		; Unsupported function
	je		.error
	cmp		ah, 0x80		; Invalid command
	je		.error
	
	jcxz	.use_ax			; BIOS may not have stored it in AX/BX, or CX/DX - test if CX == 0
	mov		ax, cx
	mov		bx, dx
	
	.use_ax:
		pop		edx
		pop		ecx
		ret
		
	.error:
		mov		ax, -1
		mov		bx, 0
		
		pop		edx
		pop		ecx
		ret


%endif