bits	16
org		0x0500

jmp		word entry

[map	symbols debug/krnldr.map]

; Includes
%include "boot/include/multiboot.s"
%include "boot/include/a20.s"
%include "boot/include/gdt.s"
%include "boot/include/memory.s"
%include "boot/include/fat12.s"

; Definitions
%define IMAGE_RMODE_BASE	0x3000
%define IMAGE_PMODE_BASE	0x100000

; Multiboot Structure
multiboot_info:
istruc multiboot_structure
	at multiboot_structure.flags,				dd 0
	at multiboot_structure.memory_low,			dd 0
	at multiboot_structure.memory_high,			dd 0
	at multiboot_structure.boot_device,			dd 0
	at multiboot_structure.cmd_line, 			dd 0
	at multiboot_structure.mods_count,			dd 0
	at multiboot_structure.mods_addr,			dd 0
	at multiboot_structure.syms0,				dd 0
	at multiboot_structure.syms1,				dd 0
	at multiboot_structure.syms2,				dd 0
	at multiboot_structure.mmap_length,			dd 0
	at multiboot_structure.mmap_addr,			dd 0
	at multiboot_structure.drives_length,		dd 0
	at multiboot_structure.drives_addr,			dd 0
	at multiboot_structure.config_table,		dd 0
	at multiboot_structure.bootloader_name,		dd 0
	at multiboot_structure.apm_table,			dd 0
	at multiboot_structure.vbe_control_info,	dd 0
	at multiboot_structure.vbe_mode_info,		dd 0
	at multiboot_structure.vbe_mode,			dw 0
	at multiboot_structure.vbe_interface_seg,	dw 0
	at multiboot_structure.vbe_interface_off,	dw 0
	at multiboot_structure.vbe_interface_len,	dw 0
iend

; Definitions
; TODO: Add Definitions Here

entry:
	cli
	mov		ax, 0x0000
	mov		ds, ax
	mov		es, ax
	mov		fs, ax
	mov		gs, ax
	
	mov		ss, ax
	mov		sp, 0xFFFF
	
	push	dx			; Save DL (boot device)
	
	; Our FAT16 bootloader (nicknamed 'DarkBoot') has now booted into KRNLDR.BIN
	; Here, we are now responsible for loading KERNEL.BIN, which is our compiled
	; C / Assembly code that will take control and setup our Operating System
	
	; KRNLDR.BIN is responsible for enabling the A20 gate to access 4GB of memory,
	; setup our first GDT so that we can enter Protected (32-bit) Mode, and then 
	; create a memory map for our system.
	
	; The last things KRNLDR.BIN is responsible for is setting up our Multiboot Information
	; This is then passed to the kernel so that it can know things like the memory map info,
	; boot device, and other information.
	
	; First, what we want to do is enable the A20 line so that we can access 4GB of memory
	call	enable_a20
	
	mov		si, MSG_A20
	call	print_msg
	
	; Now the A20 Gate is enabled and the multiboot info is set up. Our next step is to initialize 
	; the Global Descriptor Table to tell the processor what memory it can access and how
	call	install_gdt
	
	mov		si, MSG_GDT
	call	print_msg
	
	; Restore DL (boot device)
	pop		dx
	
	; Now we want to set up multiboot information. The first thing to do is set the various
	; flags in the Multiboot Specification. We want flags 0 and 1 (0x2 - 11b) to be set.
	; Flag 0 = Memory Map Information is set
	; Flag 1 = Drive Number is set
	; Flag 6 = Memory Map Info is set
	; Flag 9 = Bootloader Name is set
	mov		dword [multiboot_info + multiboot_structure.flags], 1001000011b
	mov		dword [multiboot_info + multiboot_structure.boot_device], edx
	
	; Now we want to get the available memory on the system
	xor		eax, eax
	xor		ebx, ebx
	
	; Make sure to call the function :P
	call	get_memory_size_16_mb
	
	mov		word [multiboot_info + multiboot_structure.memory_high], bx		; memory_high is the memory above 16 MB
	mov		word [multiboot_info + multiboot_structure.memory_low], ax		; memory_low is the memory between 1 MB and 16 MB
	
	; Now we want to get the memory map. Remember that this will contain all data
	; related to what memory we can use, and what memory is reserved
	mov		eax, 0
	mov		es, ax
	mov		di, 0x1000
	call	get_memory_map
	
	; On return, ES:DI is the address, BP is the count of entries
	mov		word [multiboot_info + multiboot_structure.mmap_length], bp		; The number of entries
	mov		word [multiboot_info + multiboot_structure.mmap_addr], 0x1000	; The address of entries
	
	mov		si, MSG_MEMORY
	call	print_msg
	
	; At this point, our GDT is set up, the A20 gate has been enabled, and we have loaded our
	; multiboot structure with data relating to the operating system. Our next task is to find and
	; boot into our kernel. In order to do that, we must once again search through the root directory and
	; find the data we need. Thus, we must load the FAT12 driver again.
	call	load_root_dir
	
	; Now fetch our kernel file if it exists
	mov		ebx, 0
	mov		bp, IMAGE_RMODE_BASE
	mov		si, KernelName
	
	call	load_file
	
	mov		dword [KernelSize], ecx
	cmp		ax, 0
	je		enter_protected_mode
	
	mov		si, MSG_FAILURE
	call	print_msg
	
	xor		ax, ax
	int		0x16
	int		0x19
	
	cli
	hlt
	
enter_protected_mode:
	; So now we have set up the Global Descriptor Table, enabled the A20 Gate, setup the memory map,
	; setup the Multiboot Information, and loaded our kernel (KERNEL.BIN) into memory at 0x3000.
	; The next steps are fairly simple:
	; 
	;  1. Turn off floppy driver motor (so that it does not wear itself out)
	;  2. Enter protected mode (32-bit mode) by setting PM Bit of CR0
	;  3. Copy the kernel from 0x3000 to 1 MB (0x100000)
	;  4. Far jump to the kernel and let it take control from there
	
	; Turn off floppy drive motor
	mov		dx, 0x3F2
	mov		al, 0x0C
	out		dx, al
	
	; Tell the user we are about to enter protected mode
	mov		si, MSG_BOOTING
	call	print_msg
	
	; Clear interrupts
	cli
	
	; Switch to Protected Mode
	mov		eax, cr0
	or		eax, 1
	mov		cr0, eax
	
	; Jump to our 32-bit code
	jmp		dword CODE_DESC:protected_mode

	cli
	hlt
	
; Functions

; ---------
; print_msg
; Prints a message to the console
; In:
;  DS:SI => The message to print
; ---------
print_msg:
	lodsb
	cmp		al, 0x00
	je		.done
	mov		ah, 0x0e
	int		0x10
	jmp		print_msg
	
	.done:
		ret
		
; Strings
MSG_A20		db "[DARKBOOT]: Enabled A20 Gate",10,13,0
MSG_GDT		db "[DARKBOOT]: Setup Global Descriptor Table",10,13,0
MSG_MEMORY	db "[DARKBOOT]: Loaded Available Memory and Setup Memory Map",10,13,0
MSG_BOOTING	db "[DARKBOOT]: Entering Protected Mode and Loading Kernel...",10,13,0
MSG_FAILURE	db "[DARKBOOT]: Failed to locate KERNEL.BIN. Press any key to reboot...",10,13,0

; Kernel
KernelName	db "KERNEL  BIN"
KernelSize	dd 0x00000000





; -----------
; 32 bit mode
; -----------
bits	32

protected_mode:

	; The kernel's entry point will be kernel_entry.s located in 'kernel/'. This will, in turn,
	; push the multiboot info address onto the stack so that our main() function can access the data
	; when it enters the C code kernel.
	; Thus, we must make sure to do a couple of things here:
	; 
	;  1. Copy the kernel image to 1 MB (0x100000)
	;  2. Set AX to 0x2BADB002
	;  3. Set EBX to the location of the multiboot information
	;  4. Set CS to a code segment and DS, ES, FS, GS, and SS to a valid data segment
	;  5. Set EDX to our kernel size (in sectors)
	
	; First, set up the segments
	mov		ax, DATA_DESC
	mov		ds, ax
	mov		es, ax
	mov		fs, ax
	mov		gs, ax
	
	mov		esp, 0x9000
	
copy_kernel_image:
	; Next, we need to copy the kernel image
	mov		eax, dword [KernelSize]
	movzx	ebx, word [BytesPerSector]
	mul		ebx		; KernelSize (in sectors) * BytesPerSector = Kernel Size (in bytes)
	mov		ebx, 4
	div		ebx		; Kernel Size (in bytes) / 4 = Kernel Size (dwords)
	cld
	
	mov		esi, IMAGE_RMODE_BASE
	mov		edi, IMAGE_PMODE_BASE
	mov		ecx, eax
	rep		movsd	; Copy the number of dwords in ECX from ESI to EDI

enter_kernel:
	
	; Now set the magic number and locations, and jump to our new kernel
	mov		eax, 0x2BADB002			; Set EAX to 0x2BADB002
	mov		ebx, multiboot_info		; Set EBX to the location of the multiboot information
	mov		edx, dword [KernelSize]	; Our own (internal) specification - EDX = Kernel Size (in sectors)

	jmp		CODE_DESC:IMAGE_PMODE_BASE		; Now jump to our kernel
	add		esp, 4

	cli
	hlt