#include "multiboot.h"
#include "util/types.h"
#include "video/video.h"

// Core
#include "cpu.h"
#include "gdt.h"
#include "idt.h"
#include "interrupts.h"
#include "pic.h"
#include "pit.h"

// Memory
#include "memory/pmm.h"
#include "memory/memory_map.h"

char * memory_types[] = { 
	"Available",
	"Reserved",
	"ACPI Reclaim",
	"ACPI NVS Memory"
};


void main(multiboot_t * mboot) {
	
	// So the first thing we want to do here is get the Kernel Size (in sectors) from EDX
	uint32_t kernelSize = 0;
	__asm__ volatile ("movl %%edx, %0" : "=r"(kernelSize));
	
	// Clear the screen so that we can begin printing messages
	clear_screen();
	
	printf("[DarkOS]: Welcome to DarkOS!\n");
	printf("[DarkOS]: Kernel Size is %d sectors (%d bytes)\n", kernelSize, kernelSize * 512);
	printf("[DarkOS]: Beginning Initialization...\n\n");
	
	// First, we must initialize our Global Descriptor Tables again
	gdt_init();
	printf("[DarkOS]: Initialized Global Descriptor Tables\n");
	
	// Now we want to initialize our interrupts - the ISRs and IRQs
	idt_init();
	printf("[DarkOS]: Initialized Interrupt Descriptor Table\n");
	
	pic_init();
	printf("[DarkOS]: Initialized PIC and remapped IRQs\n");
	
	interrupts_init();
	printf("[DarkOS]: Initialized Interrupts\n");
	
	pit_init(50);
	printf("[DarkOS]: Initializing PIT\n");
	
	printf("[DarkOS]: Enabling Interrupts...\n");
	__asm__("sti");
	
	printf("[DarkOS]: Finished Initializing DarkOS Core!\n\n");
	
	
	// MARK: Initialize Memory now
	printf("[DarkOS]: Initializing DarkOS Memory Management...\n");
	
	// NOTE: For some reason, this doesn't work.
	uint32_t memSize = 1024 + mboot->memory_low + mboot->memory_high * 64;
	pmm_init(memSize, 0x100000 + kernelSize * 512);
	
	printf("[DarkOS]: Initialized PMM with %i KB of Physical Memory\n", memSize);
	
	
	
	// Halt the system so that the OS never "returns" but rather waits.
	// This MUST be done so that the OS can detect interrupts when we 
	// initialize the Interrupt Vector Table (IVT)
	for (;;) {
		__asm__ volatile ("hlt");
	}
}