#include "pmm.h"
#include "util/util.h"
#include "video/video.h"

// Static variables
static uint32_t pmm_memory_size = 0;
static uint32_t pmm_used_blocks = 0;
static uint32_t pmm_max_blocks	= 0;
static uint32_t * pmm_bitmap 	= 0;

// Bitmap
void pmm_bitmap_set(uint32_t bit) {
	pmm_bitmap[bit / 32] |= (1 << (bit % 32));
}

void pmm_bitmap_clear(uint32_t bit) {
	pmm_bitmap[bit / 32] &= ~(1 << (bit % 32));
}

uint32_t pmm_bitmap_test(uint32_t bit) {
	return pmm_bitmap[bit / 32] & (1 << (bit % 32));
}

uint32_t pmm_first_free(void) {
	for (uint32_t i = 0; i < pmm_max_blocks / 32; i++) {
		if (pmm_bitmap[i] != 0xFFFFFFFF) {
			for (int j = 0; j < 32; j++) {
				int bit = 1 << j;
				if (!(pmm_bitmap[i] & bit)) {
					return i * 4 * 8 + j;
				}
			}
		}
	}
	
	return -1;
}

uint32_t pmm_first_free_s(size_t size) {
	if (size == 0) return -1;
	if (size == 1) return pmm_first_free();
	
	for (uint32_t i = 0; i < pmm_max_blocks / 32; i++) {
		if (pmm_bitmap[i] != 0xFFFFFFFF) {
			for (uint32_t j = 0; j < 32; j++) {
				int bit = 1 << j;
				if (!(pmm_bitmap[i] & bit)) {
					int startingBit = i * 32;
					startingBit += bit;
					
					uint32_t free = 0;
					for (uint32_t count = 0; count <= size; count++) {
						if (!pmm_bitmap_test(startingBit + count)) {
							free++;
						}
						
						if (free == size) {
							return i * 4 * 8 + j;
						}
					}
				}
			}
		}
	}
	
	return -1;
}

void pmm_init(uint32_t size, uint32_t bitmap) {
	pmm_memory_size	= size;
	pmm_bitmap		= (uint32_t *)bitmap;
	pmm_max_blocks	= (pmm_memory_size * 1024) / PMM_BLOCK_SIZE;
	pmm_used_blocks	= pmm_max_blocks;

	// By default, all memory is in use
	printf("%x | size: %x\n", pmm_bitmap, pmm_max_blocks / PMM_BLOCKS_PER_BYTE);
	memset((void *)pmm_bitmap, 0xF, (pmm_max_blocks / PMM_BLOCKS_PER_BYTE));
}

void pmm_init_region(uint32_t base, size_t size) {
	int align = base / PMM_BLOCK_SIZE;
	int blocks = size / PMM_BLOCK_SIZE;
	
	for (; blocks >= 0; blocks--) {
		pmm_bitmap_clear(align++);
		pmm_used_blocks--;
	}
}

void pmm_deinit_region(uint32_t base, size_t size) {
	int align = base / PMM_BLOCK_SIZE;
	int blocks = size / PMM_BLOCK_SIZE;
	
	for (; blocks >= 0; blocks--) {
		pmm_bitmap_set(align++);
		pmm_used_blocks++;
	}
}

void * pmm_alloc_block(void) {
	if (pmm_max_blocks - pmm_used_blocks <= 0) return 0;
	
	int frame = pmm_first_free();
	if (frame == -1) return 0;		// Out of memory
	
	
	pmm_bitmap_set(frame);
	
	uint32_t addr = frame * PMM_BLOCK_SIZE;
	pmm_used_blocks++;
	
	return (void *)addr;
}

void pmm_free_block(void * p) {
	uint32_t addr = (uint32_t)p;
	int frame = addr / PMM_BLOCK_SIZE;
	
	pmm_bitmap_clear(frame);
	pmm_used_blocks--;
}

void * pmm_alloc_blocks(size_t size) {
	if (pmm_max_blocks - pmm_used_blocks <= size) return 0;	// Not enough space
	
	int frame = pmm_first_free_s(size);
	if (frame == -1) return 0;		// Not enough space
	
	for (size_t i = 0; i < size; i++) {
		pmm_bitmap_set(frame + i);
	}
	
	uint32_t addr = frame * PMM_BLOCK_SIZE;
	pmm_used_blocks += size;
	
	return (void *)addr;
}

void pmm_free_blocks(void * p, size_t size) {
	uint32_t addr = (uint32_t)p;
	int frame = addr / PMM_BLOCK_SIZE;
	
	for (size_t i = 0; i < size; i++) {
		pmm_bitmap_clear(frame + i);
	}
	
	pmm_used_blocks -= size;
}